//see http://www.arduino.cc/cgi-bin/yabb2/YaBB.pl?num=1244603940
#include <WProgram.h> 

typedef unsigned char byte;
typedef unsigned int word;

#define lowByte(_w) ((_w)&0xff)
#define highByte(_w) ((_w)&0xff00)

// MCP_23017 defines
#define MCP23017_DEFAULT_ADDRESS 0b0100000	
#define MCP23017_ADDRESS_MASK    0x7 // address = 000;  7 bits!

typedef enum
{
  MCP23017_IOCON =  0x0A,
  MCP23017_IODIRA = 0x00,
  MCP23017_IODIRB = 0x01,
  MCP23017_GPIOA = 0x12,
  MCP23017_GPIOB = 0x13,
} 
MCP23017_REGISTERS;

/*
 #define MCP23017_IOCON  0x0A
 #define MCP23017_IODIRA 0x00
 #define MCP23017_IODIRB 0x01
 #define MCP23017_GPIOA  0x12
 #define MCP23017_GPIOB  0x13
 
 	#define MCP23017_GPPUA    0X0C	//pullups
 	#define MCP23017_GPINTENA 0x04
 	#define MCP23017_DEFVALA  0x06
 	#define MCP23017_INTCONA  0x08
 	#define MCP23017_INTFA    0x0E
 	#define MCP23017_INTCAPA  0x10
 	#define MCP23017_IPOLA    0x02
 */

/*
 * MCP23017 - implement a driver for the MCP 23017 port expander
 *
 * The idea is to abstract the MCP 23017 devices, which can reside on the same I2C and hide the 2 byte banks each has
 * and use the notion of offset - that is the user specify which bytes needs to be used, and the class is translating
 * to the right device (i2c address) and the right bank (GPIOA or GPIOB)
 */
class MCP23017
{

public:
  void SetDir(byte offset, byte inout_mask);//Determine in (1) or out (0) for every bit in byte
  void WriteByte (byte offset, byte i2cdata) ;   // Store 8 bits in using MCP23017
  byte ReadByte (byte offset);  // Read 8 bits from MCP23017

private:
  byte Offset2GpioRegister(byte offset);
  void WriteRegisterByte (byte i2caddress, byte i2cregister,  byte i2cdata) ;   // Store 8 bits in specified register of MCP23017
  void WriteRegisterWord (byte i2caddress, byte i2cregister,  word i2cdata);    // Store 16 bits in specified register of MCP23017
  byte ReadRegisterByte (byte i2caddress, byte i2cregister);  // Read 8 bits from MCP23017
  word ReadRegisterWord (byte i2caddress, byte i2cregister);  // Read 16 bits from MCP23017
};

/*********************************  PUBLIC ***********************************/
void MCP23017::SetDir(byte offset, byte inout_mask)
{
  byte i2caddress = offset >> 1;
  byte i2cregister;
  if ((offset & 1)== 0)
    i2cregister= MCP23017_IODIRA;
  else
    i2cregister= MCP23017_IODIRB;
  WriteRegisterByte(i2caddress, i2cregister, inout_mask); 
}

void MCP23017::WriteByte (byte offset, byte i2cdata)    // Store 8 bits in using MCP23017
{
  byte i2caddress = offset >> 1;
  byte i2cregister = Offset2GpioRegister(offset);
  WriteRegisterByte(i2caddress, i2cregister, i2cdata);
}

byte MCP23017::ReadByte (byte offset)  // Read 8 bits from MCP23017
{
  byte i2caddress = offset >> 1;
  byte i2cregister = Offset2GpioRegister(offset);
  return ReadRegisterByte (i2caddress, i2cregister);  // Read 8 bits from MCP23017  
}


/********************************* PRIVATE ***********************************/
byte MCP23017::Offset2GpioRegister(byte offset)
{
  if ((offset & 1) == 0)
    return MCP23017_GPIOA;
  else
    return MCP23017_GPIOB;
}

void MCP23017::WriteRegisterByte (byte i2caddress, byte i2cregister,  byte i2cdata)    // Store 8 bits in specified register of MCP23017
{
  byte controlbyte  =  MCP23017_DEFAULT_ADDRESS  | (i2caddress & MCP23017_ADDRESS_MASK);  // note that bit 0 not used b/c 7-bit opcode for Wire library
  Wire.beginTransmission(controlbyte);
  Wire.send(i2cregister);
  Wire.send(i2cdata);
  Wire.endTransmission();
}

void MCP23017::WriteRegisterWord (byte i2caddress, byte i2cregister,  word i2cdata)    // Store 16 bits in specified register of MCP23017
{
  byte controlbyte  =  MCP23017_DEFAULT_ADDRESS  |  (i2caddress & MCP23017_ADDRESS_MASK);  // note that bit 0 not used b/c 7-bit opcode for Wire library

  Wire.beginTransmission(controlbyte);
  Wire.send(i2cregister);
  Wire.send((byte)lowByte(i2cdata));
  Wire.send((byte)highByte(i2cdata));
  Wire.endTransmission();
}

byte MCP23017::ReadRegisterByte (byte i2caddress, byte i2cregister)  // Read 8 bits from MCP23017
{
  byte controlbyte  =  MCP23017_DEFAULT_ADDRESS  |  (i2caddress & MCP23017_ADDRESS_MASK);  // note that bit 0 not used b/c 7-bit opcode for Wire library
  byte tempdata;

  Wire.beginTransmission(controlbyte);
  Wire.send(i2cregister);
  Wire.endTransmission();

  Wire.requestFrom(controlbyte,(byte)1);  // otherwise function gets one byte and one int & chokes
  while (Wire.available() < 1) { 
  }
  tempdata = Wire.receive();      // get byte

    return tempdata;
}

word MCP23017::ReadRegisterWord (byte i2caddress, byte i2cregister)  // Read 16 bits from MCP23017
{
  byte controlbyte  =  MCP23017_DEFAULT_ADDRESS  |  (i2caddress & MCP23017_ADDRESS_MASK);  // note that bit 0 not used b/c 7-bit opcode for Wire library
  word tempdata;

  Wire.beginTransmission(controlbyte);
  Wire.send(i2cregister);
  Wire.endTransmission();

  Wire.requestFrom(controlbyte,(byte)2);  // otherwise function gets one byte and one int & chokes
  while (Wire.available() < 2) { 
  }
  tempdata = Wire.receive();      // get lower 8 bytes
  tempdata |= (Wire.receive() << 8);  // get upper 8 bytes

  return tempdata;
}








